#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2014-2016 Jan Fuchs <fuky@asu.cas.cz>
#
# Depends: libpython2.7-stdlib, libpython2.7-minimal, python-gpgme, python-pycurl
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import io
import sys
import urllib
import urllib2
import pycurl
import logging
import codecs
import smtplib
import ConfigParser
import traceback
import gpgme
import bz2
import tarfile
import hashlib
import platform
import shutil
import urlparse
import contextlib

from distutils.version import LooseVersion
from logging.handlers import TimedRotatingFileHandler

from email.header import Header
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

script_path = os.path.dirname(os.path.abspath(__file__))
logger = logging.getLogger("update_package")

class NoRedirection(urllib2.HTTPErrorProcessor):
    def http_response(self, request, response):
        return response
    https_response = http_response

class UpdatePackage:
    def __init__(self, rcp):
        self.read_cfg(rcp)
    
    def read_cfg(self, rcp):
        self.cfg = dict(rcp.items("DEFAULT"))
        self.cfg["regex_dir"] = re.compile(self.cfg["pattern_dir"])
        self.cfg["regex_package"] = re.compile(self.cfg["pattern_package"])
        self.cfg["change_destination_dir"] = bool(int(self.cfg["change_destination_dir"]))

        self.mail = dict(rcp.items("mail"))
        self.version = dict(rcp.items("version"))

    def start(self):
        ## WARNING: Package python-requests from Wheezy does not support argument "stream".
        ## argument "stream=True" - only get URL (does not download big package)
        #with contextlib.closing(requests.get(self.cfg["pkg_url"], stream=True)) as response:
        #    location = response.history[-1].headers["location"]
        #    logger.info("location = %s" % location)

        opener = urllib2.build_opener(NoRedirection)
        with contextlib.closing(opener.open(self.cfg["pkg_url"])) as response:
            location = response.info().getheader("Location")
            logger.info("location = %s" % location)

        self.cfg["pkg_url"] = location
        self.cfg["pkg_checksum_url"] = urlparse.urljoin(self.cfg["pkg_url"], self.cfg["pkg_checksum_url"])
        self.cfg["pkg_signature_url"] = urlparse.urljoin(self.cfg["pkg_url"], self.cfg["pkg_signature_url"])

        self.update = False
        self.pkg_link = "%s/%s" % (self.cfg["pkg_dir"], self.cfg["pkg_name"])
        self.installed_version = self.get_installed_version()
        self.remote_version = self.get_remote_version()

        logger.info("Installed version = %s" % (self.installed_version))
        logger.info("Remote version = %s" % (self.remote_version))

        if (self.installed_version is None):
            self.update = True
        elif (LooseVersion(self.installed_version) < LooseVersion(self.remote_version)):
            self.update = True

        if (self.update):
            self.run_update()
        else:
            logger.info("%s is up to date" % self.cfg["pkg_name"])

    def run_update(self):
        files = {
            "checksum": self.cfg["pkg_checksum_url"],
            "signature": self.cfg["pkg_signature_url"],
            "package": self.cfg["pkg_url"],
        }

        bytes_io = {}
        for key in files.keys():
            bytes_io[key] = io.BytesIO()

            logger.info("curl '%s'" % (files[key]))
            c = pycurl.Curl()
            c.setopt(c.URL, files[key].encode("ascii"))
            c.setopt(c.WRITEFUNCTION, bytes_io[key].write)
            c.perform()

            bytes_io[key].seek(0)

            http_code = c.getinfo(pycurl.HTTP_CODE)
            if (http_code != 200):
                raise Exception("HTTP code = %s" % http_code)

        ctx = gpgme.Context()
        sigs = ctx.verify(bytes_io["signature"], bytes_io["checksum"], None)[0]

        #if (sigs.summary not in [0, gpgme.SIGSUM_KEY_EXPIRED]):
        if (sigs.summary != 0):
            raise Exception("BAD signature (summary = %s)" % sigs.summary)
        if (sigs.fpr != self.cfg["fingerprint"]):
            raise Exception("BAD fingerprint '%s'" % sigs.fpr)

        # [:-8] remove ".tar.bz2"
        # firefox-46.0.tar.bz2 => firefox-46.0.en-US.linux-x86_64.tar.bz2
        filename = "%s%s" % (os.path.basename(self.cfg["pkg_url"])[:-8], self.cfg["checksum_fn_suffix"])
        bytes_io["checksum"].seek(0)
        for line in bytes_io["checksum"].readlines():
            checksum_hash, checksum_filename = line.strip().split(" ", 1)
            checksum_filename = checksum_filename.strip()
            if (checksum_filename == filename):
                break
        else:
            raise Exception("CHECKSUM not found (filename = %s)" % filename)

        package_hash = hashlib.sha512(bytes_io["package"].getvalue()).hexdigest()
        if (checksum_hash != package_hash):
            raise Exception("BAD checksum")

        bytes_io["package"].seek(0)
        bytes_io["tar"] = io.BytesIO(bz2.decompress(bytes_io["package"].read()))

        tar = tarfile.open(fileobj=bytes_io["tar"])
        pkg_name_version = "%s-%s" % (self.cfg["pkg_name"], self.remote_version)

        members = None
        if (self.cfg["change_destination_dir"]):
            members = tar.getmembers()
            idx = members[1].name.find("/")
            for member in members:
                member.name = "%s%s" % (pkg_name_version, member.name[idx:])
        else:
            dst_dir = "/".join([self.cfg["pkg_dir"], self.cfg["pkg_name"]])
            if os.path.isdir(dst_dir):
                logger.info("rmtree '%s'" % dst_dir)
                shutil.rmtree(dst_dir)

        tar.extractall(path=self.cfg["pkg_dir"], members=members)

        if (os.path.islink(self.pkg_link)):
            logger.info("Remove symbolic link %s" % self.pkg_link)
            os.unlink(self.pkg_link)

        if (self.cfg["change_destination_dir"]):
            src = "/".join([self.cfg["pkg_dir"], pkg_name_version])
            logger.info("Create symbolic link '%s' '%s'" % (src, self.pkg_link))
            os.symlink(src, self.pkg_link)

        msg = "%s installation successful" % pkg_name_version
        self.send_mail(self.mail["to"], msg)
        logger.info(msg)

    def get_installed_version(self):
        if (self.version["file_cfg"]):
            filename = "/".join([self.cfg["pkg_dir"], self.cfg["pkg_name"], self.version["file_cfg"]])
            if not os.path.exists(filename):
                return None

            logger.info("Read version from '%s'" % filename)
            rcp = ConfigParser.SafeConfigParser()
            rcp.read(filename)
            return rcp.get(self.version["section"], self.version["option"])

        if not os.path.exists(self.pkg_link):
            return None
        if not os.path.islink(self.pkg_link):
            raise Exception("'%s' is not a symbolic link" % self.pkg_link)

        dirname = os.path.basename(os.readlink(self.pkg_link))
        logger.debug("Match '%s' '%s'" % (self.cfg["pattern_dir"], dirname))

        m = self.cfg["regex_dir"].match(dirname)
        if (m is not None):
            return m.group("version")

        return None

    def get_remote_version(self):
        result = None

        url = urllib.unquote(self.cfg["pkg_url"])
        logger.debug("Match '%s' '%s'" % (self.cfg["pattern_package"], url))

        m = self.cfg["regex_package"].match(url)
        if (m is not None):
            result = m.group("version")

        if (result is None):
            raise Exception("Remote version not found")

        return result

    def send_mail(self, mail_to, text):
        if (self.mail["enable"] == "0"):
            return

        msg = MIMEMultipart("alternative")

        msg["From"] = self.mail["from"]
        msg["To"] = mail_to

        subject = "%s: %s" % (platform.node(), self.mail["subject"])
        msg["Subject"] = Header(subject, "utf-8")

        plain_text = MIMEText(text, "plain", "utf-8") 
        msg.attach(plain_text)

        smtp = smtplib.SMTP("smtp.stel")
        smtp.sendmail(self.mail["from"], mail_to.split(", "), msg.as_string())
        smtp.quit()

def main():
    filename_cfg = "update_package.cfg"
    if (len(sys.argv) == 2):
        filename_cfg = sys.argv[1]

    rcp = ConfigParser.SafeConfigParser()
    rcp.readfp(codecs.open("%s/%s" % (script_path, filename_cfg), "r", "utf-8"))
    pkg_name = rcp.get("DEFAULT", "pkg_name")

    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(name)s[%(process)d] %(threadName)s[%(thread)d] - %(levelname)s - %(message)s")

    #if (0):
    if (1):
        fh = TimedRotatingFileHandler("%s/update_package_%s.log" % (script_path, pkg_name), \
            when='D', interval=1, backupCount=30)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    else:
        sh = logging.StreamHandler(sys.stdout)
        sh.setLevel(logging.DEBUG)
        sh.setFormatter(formatter)
        logger.addHandler(sh)

    try:
        update_package = UpdatePackage(rcp)
        update_package.start()
    except Exception:
        logger.exception("exception")
        update_package.send_mail(update_package.mail["to"], traceback.format_exc())

if __name__ == '__main__':
    main()
